"""This module contains the classes that are exported from sorting package."""

from .bubblesort import BubbleSort
from .selectionsort import SelectionSort
from .insertionsort import InsertionSort
from .mergesort import MergeSort
from .quicksort import QuickSort



"""Linear search module in searching package.

The module is used to demonstrate the working of linear search algorithm

Exported methods:
    search
    evaluate
    visualize
    info
    code

Helper methods:
    __search_helper
    __print_steps

Example usage:
    search_object = searching.LinearSearch(list)
    search_object.search(7)
"""
import copy
import os
import sys

from rich.console import Console
from rich.table import Table

from ._base_class_search import BaseClass
from ._timer import Timer
from ._animate_search import _AnimateLinearSearch


class LinearSearch(BaseClass):
    """LinearSearch class which contains methods for analyzing linear search.

    Attributes:
        _datalist (list): List of ints provided by the user
    """

    def __init__(self, datalist):
        """Initializes LinearSearch class with datalist.
        Args:
            datalist (list): The list provided by the user"""
        super().__init__(datalist)
        self._datalist = datalist

    def __search_helper(self, number, steps):
        """Helper method to perform search.

        It makes a deepcopy of the datalist to prevent overrides. If the optional
        steps arguemt is true it returns a dict of all the iterations else it returns
        a f-string in 'rich' format

        Args:
            number (int): The number that we have to search
            steps (bool): bool to check if we have to print number of steps

        Returns:
            If steps is true, a dictionary is retured else a string is returned

            search_dict (dict): A dictionary of iterations mapped to where
                                search is in that iteration
            search_str (str): f-string in rich format that is printed to the
                              console
        """
        search_list = copy.deepcopy(self._datalist)

        if steps:
            search_dict = {}
            string_num = ""
            for index, value in enumerate(search_list):
                string_num = string_num + str(value) + " "
                search_dict[index] = string_num
                if value == number:
                    search_dict["[bold green] FOUND [/bold green]"] = f"[bold green]FOUND {number} at index {index}[/bold green]"
                    break

                elif index == len(search_list) - 1 and search_list[index] != number:
                    search_dict["[bold red]NOT FOUND[/bold red]"] = f"[bold red] {number} NOT FOUND in this list [/bold red]"
                    break

            return search_dict

        else:
            search_str = ""
            for index, value in enumerate(search_list):
                if value == number:
                    search_str = f"[bold green]FOUND {number} at index {index}[/bold green]"
                    break

            if not search_str:
                search_str = f"[bold red]{number} NOT FOUND in this list [/bold red]"

            return search_str

    def __print_steps(self, step_dict):
        """Helper method to print the 'search_dict' to console

        Steps are printed using the rich.table module

        Args:
            step_dict (dict): The dict to be printed
        """

        table = Table(title=" Linear search steps")
        table.add_column("At index", justify="center", style="cyan")
        table.add_column("List", style="magenta", no_wrap=False)

        for iteration, step_list in step_dict.items():
            table.add_row(str(iteration), str(step_list))

        console = Console()
        console.print(table)

    def search(self, number, steps=False):
        """Searches if number is in the list and prints the result to the console.

        Set optional parameter 'steps' to True if you want to print the iterations
        table to console

        Args:
            number (int): The number to be searched
            steps (bool): Optional; (default: False)
        """

        # _search_result can be either dict or str
        _search_result = self.__search_helper(number, steps)

        # if steps is True, means it's a dict and we pass it to __print_steps
        # else we just print it
        if steps:
            return self.__print_steps(_search_result)
        console = Console()
        return console.print(_search_result)

    def evaluate(self, number, iterations=1):
        """Performs linear search on the list and prints the result to the
        console.

        Set optional parameter 'iterations' to the number of times you want to
        perform linear search on the list

        Args:
            number (int): The number to be searched
            iterations (int): Optional; (default: 1)

        Raises:
            TypeError: When user inputs anything other than an int for number
                       or iteration
        """
        if not isinstance(number, int):
            raise TypeError('Number can only be int datatype')

        if not isinstance(iterations, int):
            raise TypeError('Iterations can only be int datatype')

        _eval_list = copy.deepcopy(self._datalist)
        _length_of_list = len(_eval_list)
        _timing_list = []

        _eval_iter = iterations

        while _eval_iter:
            # instantiating the timer
            timer = Timer()
            timer.start()

            # stopping it only when number is found or we have reached end
            # of the list and havent found the number
            for _index, _value in enumerate(_eval_list):
                if _value == number:
                    _stop = timer.stop()
                    _timing_list.append(_stop)
                    break
                elif _index == _length_of_list - 1:
                    _stop = timer.stop()
                    _timing_list.append(_stop)
                    break

            _eval_iter -= 1

        _minimum_time = min(_timing_list)
        _maximum_time = max(_timing_list)
        _average_time = int(sum(_timing_list) / iterations)

        _eval_dict = {
            "Minimum Time": _minimum_time,
            "Maximum Time": _maximum_time,
            "Average Time": _average_time
        }

        # print statements for better formatting
        print()
        self.search(number, steps=False)
        print()

        # calling the function to print the dictionary
        return super()._print_evaluate(_eval_dict, "Linear Search")

    def visualize(self, number, interval=100):
        """Shows a visualization using matplotlib of linear search performed on
        the list user passed.

        Set optional parameter 'interval' to change the delay between frames
        in milliseconds.

        Args:
            number (int): The number to be searched
            interval (int): Optional; (default: 100)
                            Delay between frames in milliseconds

        Raises:
            TypeError: An error when user inputs anything other than an int for number
                       or interval
        """
        if not isinstance(number, int):
            raise TypeError('Number can only be int datatype')

        if not isinstance(interval, int):
            raise TypeError('Interval can only be int datatype')

        # Instantiating the AnimateLinearSearch class whose init method calls the
        # AnimateAlgorithm function which performs the animation

        # ISSUE: It gets extremely slow on values more than 25
        # any way to optimize??
        _AnimateLinearSearch(self._datalist, number, interval, "Linear Search")

    @classmethod
    def info(cls):
        """Class method that provides information on linear search."""
        dir = os.path.dirname(__file__)
        path_to_information = os.path.join(dir, '_markdown_files', 'linearsearch.md')
        super()._print_info(cls, path_to_information)

    @classmethod
    def code(cls):
        """Class method that prints the function for linear search algorithm in console."""
        ls_code = """
        # function for linear search
        # returns the index of the element to search
        # else returns -1
        def linear_search(array,x):

            for index, number in enumerate(array):
                if number == x:
                    return index

            #if number is not found, we return -1
            return -1
        """
        super()._print_code(cls, ls_code)



"""Merge sort module in sorting package.

The module is used to demonstrate the working of merge sort algorithm

Exported methods from MergeSort class:
    sort
    evaluate
    visualize
    info
    code

Helper methods:
    __sort_it
    __eval_helper
    __fast_merge_asc
    __fast_merge_desc
    __no_len_ms


Helper generators:
    __ascending_sort_merge_algo
    __descending_sort_merge_algo
    __merge_algo
    __animate_sort_it

Example usage:
    bs_object = sorting.MergeSort(datalist)
    bs_object.sort(reverse = True, steps = True)
"""

import copy
import os
import sys

from ._base_sorting import BaseClass
from ._timer import Timer
from ._animation import animate_algorithm


class MergeSort(BaseClass):
    """Merge Sort class which contains methods for analyzing merge sort algorithm.

    Attributes:
        _datalist (list): List of ints provided by the user
    """

    def __init__(self, datalist):
        """Initializes Merge Sort class with datalist.

        Args:
            datalist (list): The list provided by the user
        """
        super().__init__(datalist)
        self._datalist = datalist

    def __repr__(self):
        """__repr__ for MergeSort class"""
        return f'algovis.sorting.mergesort.MergeSort({self._datalist})'

    def __ascending_sort_merge_algo(self, passed_list, start, mid, end):
        """Helper generator for the ascending merge sort algorithm.

        It is called by __merge_algo

        Yields:
              asc_list (list): yields list after each pass of merge sort on
                               the list

        """
        merged = []
        leftIdx = start
        rightIdx = mid + 1

        while leftIdx <= mid and rightIdx <= end:
            if passed_list[leftIdx] < passed_list[rightIdx]:
                merged.append(passed_list[leftIdx])
                leftIdx += 1
            else:
                merged.append(passed_list[rightIdx])
                rightIdx += 1

        while leftIdx <= mid:
            merged.append(passed_list[leftIdx])
            leftIdx += 1

        while rightIdx <= end:
            merged.append(passed_list[rightIdx])
            rightIdx += 1

        for i, sorted_val in enumerate(merged):
            passed_list[start + i] = sorted_val

        yield passed_list

    def __descending_sort_merge_algo(self, passed_list, start, mid, end):
        """Helper generator for the descending merge sort algorithm.

        It is called by __merge_algo

        Yields:
              desc_list (list): yields list after each pass of merge sort on
                               the list

        """
        merged = []
        leftIdx = start
        rightIdx = mid + 1

        while leftIdx <= mid and rightIdx <= end:
            if passed_list[leftIdx] > passed_list[rightIdx]:
                merged.append(passed_list[leftIdx])
                leftIdx += 1
            else:
                merged.append(passed_list[rightIdx])
                rightIdx += 1

        while leftIdx <= mid:
            merged.append(passed_list[leftIdx])
            leftIdx += 1

        while rightIdx <= end:
            merged.append(passed_list[rightIdx])
            rightIdx += 1

        for i, sorted_val in enumerate(merged):
            passed_list[start + i] = sorted_val

        yield passed_list

    def __merge_algo(self, passed_list, start, end, reverse):
        """Helper method for merge sort

        It calls the __ascend... or __descen... recursively to perform
        merge sort algo on the list.

        Yields:
            passed_list (list): list on which we're performing the algo is
                                yielded after each iteration
        """
        if end <= start:
            return

        mid = start + ((end - start + 1) // 2) - 1
        yield from self.__merge_algo(passed_list, start, mid, reverse)
        yield from self.__merge_algo(passed_list, mid + 1, end, reverse)
        if not reverse:
            yield from self.__ascending_sort_merge_algo(passed_list, start, mid, end)
        else:
            yield from self.__descending_sort_merge_algo(passed_list, start, mid, end)
        yield passed_list

    def __animate_sort_it(self, reverse=False):
        """The generator passed to _animate_algorithm.

        Args:
            reverse (bool): Boolean to sort list in reverse

        Yields:
            iter_list (list): List after each iteration of Merge sort
        """
        passed_list = copy.deepcopy(self._datalist)
        for iter_list in self.__merge_algo(passed_list, 0, len(passed_list) - 1, reverse):
            yield iter_list

    def __sort_it(self, reverse, steps):
        """Helper method for 'sort' method

        It checks which generator to call based on reverse. Then it stores every
        iteration of yielded list in a dictionary. If steps is true, then _print_steps
        is called from BaseClass and the dictionary iteration_dict is passed to it which
        prints the steps

        Args:
            reverse (bool): based on reverse, we call the generator to either sort the list
                            in ascending or descending order
            steps (bool): If true, it calls function to print steps

        Returns:
            iteration_dict (dict):
        """
        passed_list = copy.deepcopy(self._datalist)
        iteration_dict = {}
        iterations = 0

        # the 0th iteration is basically the passed list
        iteration_dict[iterations] = self._datalist

        for yielded_list in self.__merge_algo(passed_list, 0, len(passed_list) - 1, reverse):
            iterations += 1
            iteration_dict[iterations] = copy.deepcopy(yielded_list)

        if steps:
            super()._print_steps(iteration_dict, "Merge Sort")

        return iteration_dict

    def __fast_merge_asc(self, array1, array2):
        """Helper method for 'evaluate'

        using https://stackoverflow.com/questions/7063697/why-is-my-mergesort-so-slow-in-python
        Merge Sort with optimizations for making it much faster

        """
        merged_array = []
        while array1 or array2:
            if not array1:
                merged_array.append(array2.pop())
            elif (not array2) or array1[-1] > array2[-1]:
                merged_array.append(array1.pop())
            else:
                merged_array.append(array2.pop())
        merged_array.reverse()
        return merged_array

    def __fast_merge_desc(self, array1, array2):
        """Helper method for 'evaluate'.

        Similar to above method, in descending order

        """
        merged_array = []
        while array1 or array2:
            if not array1:
                merged_array.append(array2.pop())
            elif (not array2) or array1[-1] < array2[-1]:
                merged_array.append(array1.pop())
            else:
                merged_array.append(array2.pop())
        merged_array.reverse()
        return merged_array

    def __no_len_ms(self, array, reverse):
        """Helper method for 'evaluate'.

        It calls __fast-merge which is optimized version of mergesort for evaluation

        Returns:
            A sorted array, which is not used anywhere
        """
        n = len(array)
        if n <= 1:
            return array
        mid = int(n / 2)
        left = array[:mid]
        right = array[mid:]
        if not reverse:
            return self.__fast_merge_asc(self.__no_len_ms(left, reverse), self.__no_len_ms(right, reverse))
        return self.__fast_merge_desc(self.__no_len_ms(left, reverse), self.__no_len_ms(right, reverse))

    def __eval_helper(self, reverse, iterations):
        """Helper method for 'evaluate' method

        Evaluating time of ascending merge sort algorithm.Takes in the
        'iterations' provided by the user and performs sorting that many times.

        Args:
            iterations (int): Number of times to perform sorting on this algo
            reverse (bool): check if user wants to reverse the list

        Returns:
            timing_list (list): A list of time taken for fully sorting the
                                list 'iterations' number of times
        """
        timing_list = []

        while iterations:
            time_list = copy.deepcopy(self._datalist)

            timer = Timer()
            timer.start()

            self.__no_len_ms(time_list, reverse)

            stop = timer.stop()
            timing_list.append(stop)

            iterations -= 1

        return timing_list

    def sort(self, reverse=False, steps=False):
        """Performs merge sort on the list provided by the user.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            steps (bool): Optional; (default: False)
                            If set to True, shows iteration of each pass of
                            mergesort on the list

        Returns:
            sorted_list (list): sorted list
        """
        _sorted_object = self.__sort_it(reverse, steps)

        sorted_list = list(_sorted_object.values())[-1]

        return sorted_list

    def evaluate(self, reverse=False, iterations=1):
        """Prints the time taken to perform merge sort in nanoseconds and
        seconds to the console.

        Set optional parameter 'iterations' to the number of times you want to
        perform merge sort on the list. After every iteration, the list is reset to
        it's original unsorted state.

        Set optional parameter 'reverse' to True to sort the list in descending order.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            iterations (int): Optional; (default: 1)
                              Number of times to perform merge sort on list

        Raises:
            TypeError: When user inputs anything other than an int for iteration
        """

        if not isinstance(iterations, int):
            raise TypeError('Iterations can only be int datatype')

        _timing_list = self.__eval_helper(reverse, iterations)

        _minimum_time = min(_timing_list)
        _maximum_time = max(_timing_list)
        _average_time = int(sum(_timing_list) / iterations)

        _eval_dict = {
            "Minimum Time": _minimum_time,
            "Maximum Time": _maximum_time,
            "Average Time": _average_time
        }

        return super()._print_evaluate(_eval_dict, "Merge Sort")

    def visualize(self, reverse=False, interval=250):
        """Shows a visualization using matplotlib of merge sort performed on
        the list user passed.

        Set optional parameter 'interval' to change the delay between frames
        in milliseconds.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            interval (int): Optional; (default: 250)
                            Delay between frames in milliseconds

        Raises:
            TypeError: An error when user inputs anything other than int for
                       interval
        """
        if not isinstance(interval, int):
            raise TypeError('Interval can only be int datatype')

        _vis_list = copy.deepcopy(self._datalist)

        animate_algorithm("Merge Sort", _vis_list, self.__animate_sort_it(reverse), interval, operations=True)

    @classmethod
    def info(cls):
        """Class method that provides information on merge sort."""
        dir = os.path.dirname(__file__)
        path_to_information = os.path.join(dir, '_markdown_files', 'mergesort.md')
        return super()._print_info(path_to_information)

    @classmethod
    def code(cls):
        """Class method that prints code to the console."""
        my_code = """
        def merge(array, left_index, right_index, middle):
            # Make copies of both arrays we're trying to merge

            # The second parameter is non-inclusive, so we have to increase by 1
            left_copy = array[left_index:middle + 1]
            right_copy = array[middle+1:right_index+1]

            # Initial values for variables that we use to keep
            # track of where we are in each array
            left_copy_index = 0
            right_copy_index = 0
            sorted_index = left_index

            # Go through both copies until we run out of elements in one
            while left_copy_index < len(left_copy) and right_copy_index < len(right_copy):

                # If our left_copy has the smaller element, put it in the sorted
                # part and then move forward in left_copy (by increasing the pointer)
                if left_copy[left_copy_index] <= right_copy[right_copy_index]:
                    array[sorted_index] = left_copy[left_copy_index]
                    left_copy_index = left_copy_index + 1
                # Opposite from above
                else:
                    array[sorted_index] = right_copy[right_copy_index]
                    right_copy_index = right_copy_index + 1

                # Regardless of where we got our element from
                # move forward in the sorted part
                sorted_index = sorted_index + 1

            # We ran out of elements either in left_copy or right_copy
            # so we will go through the remaining elements and add them
            while left_copy_index < len(left_copy):
                array[sorted_index] = left_copy[left_copy_index]
                left_copy_index = left_copy_index + 1
                sorted_index = sorted_index + 1

            while right_copy_index < len(right_copy):
                array[sorted_index] = right_copy[right_copy_index]
                right_copy_index = right_copy_index + 1
                sorted_index = sorted_index + 1

        def merge_sort(array, left_index, right_index):
            if left_index >= right_index:
                return

            middle = (left_index + right_index)//2
            merge_sort(array, left_index, middle)
            merge_sort(array, middle + 1, right_index)
            merge(array, left_index, right_index, middle)

        #from https://stackabuse.com/merge-sort-in-python/
        """
        return super()._print_code(my_code)



"""Helper animation module in sorting package.

This module contains 2 functions, animate algorithm and _update fig which are
used for creating the animations. The animate_algorithm is exported to public
sorting modules and parameters are passed to it which help in making the animation.
_update_fig is used to update the figure at every iteration/operation.

Read more about it here
https://matplotlib.org/3.2.1/api/_as_gen/matplotlib.animation.FuncAnimation.html#matplotlib.animation.FuncAnimation
"""
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.ticker import MaxNLocator


def _update_fig(passed_list, rects, iteration, text, operations):
    """Method called by animation.FuncAnimation that updates the figure for
    each frame.

        Args:
            passed_list (list): The list yielded by generator which contains heights
                                of the bars
            rects (plt.container.BarContainer): bar container object that is
                                 used to set heghts
            iteration (list): Used to keep track of iterations/operations
            text (plt.text.Text): text displayed on animation
            operations (bool): if true, it dispays operations instead of iterations
                                in the text
    """
    for rect, val in zip(rects, passed_list):
        rect.set_height(val)
    iteration[0] += 1
    if operations:
        text.set_text(f"# of operations: {iteration[0]}")
    else:
        text.set_text(f"# of iterations: {iteration[0]}")


def animate_algorithm(title, passed_list, passed_generator, interval, operations=False):
    """Method that initializes the animation

        Args:
            title (str): Heading of the animation
            passed_list (list): The list provided by the user
            passed_generator (generator): the generator used in animation to
                                          create each frame
            interval (int): delay between frames in milliseconds
            operations (bool): Optional; (default: False)
                                If set to true, it displays operations instead of
                                iterations in the animation text

        Raises:
            ValueError: If the list contains negative numbers it raises this error
    """

    for elem in passed_list:
        if elem < 0:
            raise ValueError("List cannot contain negative elements")

    plt.style.use('dark_background')

    fig, ax = plt.subplots(figsize=(13, 6.5))
    # fig.set_tight_layout(True)
    ax.set_title(title)

    # setting the x and y limits
    ax.set_xlim(0, len(passed_list))
    ax.set_ylim(0, int(1.09 * max(passed_list)))

    # making x axis labels evenly spaced and integers
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    rects = ax.bar(range(len(passed_list)), passed_list, align="edge")

    # the text that'll be changed
    text = ax.text(0.02, 0.95, "", transform=ax.transAxes)

    iteration = [0]

    anim_sorting = animation.FuncAnimation(fig, func=_update_fig,
                                           fargs=(rects, iteration, text, operations), frames=passed_generator, interval=interval,
                                           repeat=False, save_count=10000)

    # saving figures
    # Writer = animation.writers['ffmpeg']
    # writer = Writer(fps=1, metadata=dict(artist='Me'), bitrate=50000)
    # anim_sorting.save(f"{title.split()[0]}.mp4", writer=writer)

    plt.show()



"""Helper sorting module in sorting package.

This module contains the BaseClass class which all the sorting algorithms
inherit. Rich library is used to print out to the console.
"""
from rich.console import Console
from rich.table import Table
from rich.syntax import Syntax
from rich.markdown import Markdown


class BaseClass:
    """Baseclass is a class inherited by every other public sorting module

    It contains all the functions that are relevent to every sorting module.
    super().__init()__ is called by every sorting module.
    """

    def __init__(self, datalist):
        """Checks the passed object, raises errors otherwise

        Args:
            datalist (list): The list provided by the user

        Raises:
            TypeError: If object passed is not a list or doesn't contain only integers
            ValueError: If the list is empty

        """
        if not isinstance(datalist, list):
            raise TypeError('Parameter must be a list datatype')

        if not datalist:
            raise ValueError('Parameter has invalid value')

        for _element in datalist:
            if not isinstance(_element, int):
                raise TypeError('Parameter can only contain int datatype')

    def _print_steps(self, step_dict, name):
        """Helper method to print the steps.

        Uses Rich to create a table and then print it
        Args:
            step_dict (dict): Dictionary containing all the steps used to
                              create a table
            name (str): Gives title to the table
        """
        table = Table(title=f"{name} Steps")
        table.add_column("Iteration", justify="center", style="cyan")
        table.add_column("List", style="magenta", justify="center", no_wrap=False)

        for iteration, step_list in step_dict.items():
            list_x = " ".join(map(str, step_list))
            table.add_row(str(iteration), str(list_x))
        console = Console()
        console.print(table)

    def _print_evaluate(self, eval_dict, name):
        """Helper method to print the 'evaluate' dictionary

        Args:
            eval_dict (dict): The dictionary that we pass after evaluating
                              the algorithm
            name (str): gives heading to the table
        """
        table = Table(title=name + " Evaluation")
        table.add_column("Metric", justify="center", style="cyan")
        table.add_column("Time (ns)", style="magenta", justify="center", no_wrap=False)
        table.add_column("Time (sec)", style="magenta", justify="center", no_wrap=False)

        for metric, time_ns in eval_dict.items():
            table.add_row(str(metric), str(time_ns), str("{:.9f}".format(time_ns / 1e9)))

        console = Console()
        console.print(table)

    def _print_info(path_to_information):
        """Helper method to print the information in the console using rich markdown

        Args:
            path_to_information (str): path to the markdown file (in _markdown_files folder)
        """
        console = Console()
        with open(path_to_information) as readme:
            information = Markdown(readme.read(), justify="full")

        console.print(information)

    def _print_code(my_code):
        """Helper method to print the code in the console using rich syntax

        Args:
            code_str (str): (doc)string containing code
        """
        syntax = Syntax(my_code, "python", theme="monokai", line_numbers=True)
        console = Console()
        console.print(syntax)



"""Binary search module in searching package.

The module is used to demonstrate the working of binary search algorithm

Exported methods:
    search
    evaluate
    visualize
    info
    code

Helper methods:
    __search_helper
    __print_steps

Example usage:
    search_object = searching.Binary(list)
    search_object.search(42)
"""
import copy
import os
import sys

from rich.console import Console
from rich.table import Table

from ._base_class_search import BaseClass
from ._timer import Timer
from ._animate_search import _AnimateBinarySearch


class BinarySearch(BaseClass):
    """BinarySearch class which contains methods for analyzing binary search.

    Attributes:
        _datalist (list): List of ints provided by the user
    """

    def __init__(self, datalist):
        """Initializes BinarySearch class with datalist.

        Args:
            datalist (list): The list provided by the user. If the list is unsorted,
                             it is automatically sorted
        """
        super().__init__(datalist)
        self._datalist = sorted(datalist)

    def __repr__(self):
        return f'algovis.searching.binarysearch.BinarySearch({self._datalist})'

    def __search_helper(self, number, steps):
        """Helper method to perform search.

        It makes a deepcopy of the datalist to prevent accidental overrides. If the optional
        steps arguemt is true it returns a tuple of lists of list of all the iterations (multiple values)
        and index of 'number'(if present) else -1.
        Else it returns a f-string in 'rich' format

        Args:
            number (int): The number that we have to search
            steps (bool): Bool to check if we have to print number of steps

        Returns:
            If steps is true, a tuple of (list of lists, index of element) is returned else a tuple of
            (f-string,index of element) is returned. If element is not found, index is -1

            list_of_iterations, index: A tuple of (list of lists, index_of_searched_element)
            f-string, index: A tuple of (f-string, index_of_searched_element)
        """
        search_list = copy.deepcopy(self._datalist)

        if steps:
            list_of_iterations = []
            counter = 0
            left_index = 0
            right_index = len(search_list) - 1

            while left_index <= right_index:
                temp_list = []
                temp_list.append(str(counter))
                temp_list.append(str(left_index))
                temp_list.append(str(right_index))

                middle_index = left_index + (right_index - left_index) // 2

                temp_list.append(str(middle_index))

                temp_list.append(search_list[left_index:right_index + 1])

                list_of_iterations.append(temp_list)

                counter = counter + 1

                if search_list[middle_index] == number:
                    return list_of_iterations, middle_index

                elif search_list[middle_index] < number:
                    left_index = middle_index + 1

                else:
                    right_index = middle_index - 1

            return list_of_iterations, -1

        else:

            left_index = 0
            right_index = len(search_list) - 1

            while left_index <= right_index:

                middle_index = left_index + (right_index - left_index) // 2

                if search_list[middle_index] == number:
                    return f"[bold green]FOUND {number} at index {middle_index}[/bold green]", 0

                elif search_list[middle_index] < number:
                    left_index = middle_index + 1

                else:
                    right_index = middle_index - 1

            return f"[bold red]{number} NOT FOUND in this list [/bold red]", -1

    def __print_steps(self, list_of_iterations, index, number):
        """Helper method to print the list_of_iterations to console.

        Steps are printed using the rich.table module

        Args:
            list_of_iterations (list): A list of lists containing data about every iteration
            index (int): The index of element (number)
            number (int): Element to be searched

        """
        table = Table(title="Binary search steps")
        table.add_column("Iteration", justify="center", style="cyan")
        table.add_column("Left index", justify="center", style="cyan")
        table.add_column("Middle index", justify="center", style="cyan")
        table.add_column("Right index", justify="center", style="cyan")
        table.add_column("List", style="magenta", justify="center", no_wrap=False)

        for iteration in list_of_iterations:
            iter_ = str(iteration[0])
            left_ix = str(iteration[1])
            right_ix = str(iteration[2])
            middle_ix = str(iteration[3])
            list_ = " ".join(str(i) for i in iteration[4])
            table.add_row(iter_, left_ix, middle_ix, right_ix, list_)

        # adding last row based on if we found the number or not
        if index == -1:
            table.add_row(" ", " ", " ", " ", f"[bold red]{number} NOT FOUND in this list [/bold red]")
        else:
            table.add_row(" ", " ", " ", " ", f"[bold green]FOUND {number} at index {index}[/bold green]")

        console = Console()
        console.print(table)

    def search(self, number, steps=False):
        """Performs iterative binary search on the list and prints the result
        to the console.

        Set optional parameter 'steps' to True if you want to print the iterations
        table to console

        Args:
            number (int): The number to be searched
            steps (bool): Optional; (default: False)
        """

        _search_result, index = self.__search_helper(number, steps)

        if steps:
            return self.__print_steps(_search_result, index, number)
        else:
            console = Console()
            return console.print(_search_result)

    def evaluate(self, number, iterations=1):
        """Prints the time taken to perform binary search in nanoseconds and seconds
        to the console.

        Set optional parameter 'iterations' to the number of times you want to
        perform binary search on the list

        Args:
            number (int): The number to be searched
            iterations (int): Optional; (default: 1)

        Raises:
            TypeError: When user inputs anything other than an int for number
                       or iteration
        """

        if not isinstance(number, int):
            raise TypeError('Number can only be int datatype')

        if not isinstance(iterations, int):
            raise TypeError('Iterations can only be int datatype')

        _eval_list = copy.deepcopy(self._datalist)
        _timing_list = []
        _eval_iter = iterations

        while _eval_iter:
            timer = Timer()
            timer.start()

            left_index = 0
            right_index = len(_eval_list) - 1

            while left_index <= right_index:
                middle_index = left_index + (right_index - left_index) // 2

                if _eval_list[middle_index] == number:
                    _stop = timer.stop()
                    _timing_list.append(_stop)
                    break
                elif _eval_list[middle_index] < number:
                    left_index = middle_index + 1
                else:
                    right_index = middle_index - 1

            # if we don't find the number and reach the end
            if _eval_list[middle_index] != number:
                _stop = timer.stop()
                _timing_list.append(_stop)

            _eval_iter -= 1

        _minimum_time = min(_timing_list)
        _maximum_time = max(_timing_list)
        _average_time = int(sum(_timing_list) / iterations)

        _eval_dict = {
            "Minimum Time": _minimum_time,
            "Maximum Time": _maximum_time,
            "Average Time": _average_time
        }

        print()
        self.search(number, steps=False)
        print()

        return super()._print_evaluate(_eval_dict, "Binary Search")

    def visualize(self, number, interval=1000):
        """Shows a visualization using matplotlib of binary search performed on
        the list user passed.

        Set optional parameter 'interval' to change the delay between frames
        in milliseconds.

        Args:
            number (int): The number to be searched
            interval (int): Optional; (default: 1000)
                            Delay between frames in milliseconds

        Raises:
            TypeError: An error when user inputs anything other than int for number
                       or interval
        """
        if not isinstance(number, int):
            raise TypeError('Number can only be int datatype')

        if not isinstance(interval, int):
            raise TypeError('Interval can only be int datatype')

        # Instantiating the AnimateBinarySearch class whose init method calls the
        # AnimateAlgorithm function which performs the animation
        _AnimateBinarySearch(self._datalist, number, interval, "Binary Search")

    @classmethod
    def info(cls):
        """Class method that provides information on binary search."""
        dir = os.path.dirname(__file__)
        path_to_information = os.path.join(dir, '_markdown_files', 'binarysearch.md')
        super()._print_info(cls, path_to_information)

    @classmethod
    def code(cls):
        """Class method that prints the function for binary search algorithm in console."""
        bs_code = """
        # iterative binary search
        def binary_search(array, low, mid, high):

            while low <= high:

                mid = low + (high - low)//2

                # Check if x is present at mid
                if array[mid] == x:
                    return mid

                # If x is greater, ignore left half
                elif array[mid] < x:
                    low = mid + 1

                # If x is smaller, ignore right half
                else:
                    high = mid - 1

            # If we reach here, then the element
            # was not present
            return -1
        """
        super()._print_code(cls, bs_code)



"""Helper module in sorting package.

From https://realpython.com/python-timer/#a-python-timer-class
Changed a bit to return nanoseconds in int instead of seconds by
using perf_counter_ns()
"""
import time


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class Timer:
    def __init__(self):
        self._start_time = None

    def start(self):
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter_ns()

    def stop(self):
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter_ns() - self._start_time
        return elapsed_time



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['algovis', 'algovis.searching', 'algovis.sorting']

package_data = \
{'': ['*'],
 'algovis.searching': ['_markdown_files/*'],
 'algovis.sorting': ['_markdown_files/*']}

install_requires = \
['matplotlib>=3.2.1,<4.0.0', 'rich>=3.3.1,<4.0.0']

setup_kwargs = {
    'name': 'algovis',
    'version': '0.1.6',
    'description': 'A Python library for visualising algorithms',
    'long_description': '[![CodeFactor](https://www.codefactor.io/repository/github/hotshot07/algovis/badge/master?s=197e9c6e50413744c0a2c43785a6dee096ee1a4d)](https://www.codefactor.io/repository/github/hotshot07/algovis/overview/master)  ![PyPI](https://img.shields.io/pypi/v/algovis) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/algovis) [![Downloads](https://pepy.tech/badge/algovis)](https://pepy.tech/project/algovis) <!-- ![PyPI - Downloads](https://img.shields.io/pypi/dm/algovis) --> ![GitHub last commit](https://img.shields.io/github/last-commit/hotshot07/algovis)  ![PyPI - License](https://img.shields.io/pypi/l/algovis) ![Netlify](https://img.shields.io/netlify/f4cede18-f2c6-4299-abc1-92b8a8ef9995) [![Made by](https://img.shields.io/badge/Made%20by-hotshot07-blue)](https://mackweb.in) [![Twitter Follow](https://img.shields.io/twitter/follow/gamesetmatch07?style=social)](https://twitter.com/elonmusk)\n\nAlgovis is a python library made for visualizing algorithms. Refer to the [documentation](https://algovisdocs.netlify.app/) for more info.\n\nCurrently the library has these algorithms\n\n#### Sorting\n\n- Bubble Sort\n- Insertion Sort\n- Selection Sort\n- Merge Sort\n- Quick Sort\n\n#### Searching\n\n- Linear Search\n- Binary Search\n\n\n## Getting Started\n\n### Prerequisites\n\nI would highly suggest making a virtual environment. The main purpose of a Python virtual environments is to create an isolated environment for Python projects. You can read more about them [here](https://realpython.com/python-virtual-environments-a-primer/).\n\n```bash\n# making a test folder\n\n$mkdir test_algovis\n\n# make it the current directory\n\n$cd test_algovis\n\n# making a virtual environment (you can replace envname with whatever name you like)\n\n$python3 -m venv envname\n\n# activating it\n\n$source envname/bin/activate\n```\n\nYou can only access algovis inside this virtual environment. To leave this virtual env when you\'re done trying out the library, type\n\n```bash\n$deactivate\n```\n\n### Installing\n\n```bash\n$pip3 install algovis\n```\n\n\n\n### Using the sorting package\n\n\n\n#### Visualize method\n\n```python\n# import the sorting package from library\nfrom algovis import sorting\n\n# importing random module to shuffle the list\nimport random\n\n# Making a list of 100 integers from 1-100\n# using list comprehension\nmy_list = [i+1 for i in range(100)]\n\n# shuffling the list\nrandom.shuffle(my_list)\n\n# making a BubbleSort class object by passing the shuffled list\nbs_object = sorting.BubbleSort(my_list)\n\n# calling the visualize method\nbs_object.visualize(interval= 100)\n```\n##### Output\n<img src="https://media.giphy.com/media/ieb13rrmvVWC02zmI8/giphy.gif" width="600">\n\n\n\n#### sort method\n\n```python\n# lets work on a shorter example now\nmy_list = [i + 1 for i in range(10)]\n\n# shuffling the list using random module\nrandom.shuffle(my_list)\n\n#making a quicksort object\nqs_object = sorting.QuickSort(my_list)\n\n#sorting in reverse with steps\nqs_object.sort(pivot = "first", steps = True, reverse = True)\n\n```\n\n#### evaluate method\n```python\n# calling the evaluate method and passing the optional parameter \'iterations\'\n# the list is sorted \'iterations\' number of times and the min, max and average time taken\n#to sort the list is returned in form of a formatted table\nbs_object.evaluate(iterations = 100)\n```\n\n\n#### info method\n```python\n# This method gives us information about the algorithm\nbs_object.info()\n```\n\n#### code method\n```python\n# It prints out the function for the algorithm\nbs_object.code()\n```\n\n>My terminal config is iTerm2 + ohmyzsh + powerlevel10k with dark backgroud. Colors may appear different in your terminal output. It\'s recommended to change the terminal color to something darker\n\n### Using the searching package\n\n>The searching package has the same methods as sorting, just instead of \'sort\' you have \'search\'\n>Refer to the [documentation](https://algovisdocs.netlify.app/) for more info\n\n#### search method\n```python\n#importing searching package\nfrom algovis import searching\n\n# making a list of integers from 1 to 100\n# using list comprehension\nmy_list = [i+1 for i in range(100)]\n\n#making binary search object\nbin_search = searching.BinarySearch(my_list)\n\n#calling the search method\nbin_search.search(42, steps = True)\n```\n\n#### visualize method\n\n```python\n# calling the visualize method\n# interval is the time between two different frames of the animation\nbin_search.visualize(42, interval = 1000)\n```\n<img src ="https://media.giphy.com/media/l3Cktj8ULcjHK3c90E/giphy.gif" width ="600">\n\n```python\n# or if you want to linear search 42\nlin_search = searching.LinearSearch(my_list)\n\n# setting a less interval to make a much faster animation\nlin_search.visualize(42, interval = 100)\n```\n\n### Doumentation\nThe documentation is built with [MKdocs](https://www.mkdocs.org/) using [material](https://squidfunk.github.io/mkdocs-material/) theme and is hosted on netlify. You can read it [here](https://algovisdocs.netlify.app/)\n\n### Author\n\n* **Mayank Arora** *(hotshot07)*\n\n\n### License\n\nThis project is licensed under the GNU Affero General Public License v3 (AGPL-3.0) - see the [LICENSE](LICENSE) file for details',
    'author': 'hotshot07',
    'author_email': 'aroram@tcd.ie',
    'maintainer': 'hotshot07',
    'maintainer_email': 'aroram@tcd.ie',
    'url': 'https://github.com/hotshot07/algovis',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



"""Helper animation module in searching package.

This module contains 2 classes, AnimateBinarySearch and AnimateLinearSearch
which contain methods that help in producing the visulization. These classes
are exported to respective modules and when instantiated, the self method
calls the animate_algorithm function(each class has it's own).

Read more about FuncAnimation here
https://matplotlib.org/3.2.1/api/_as_gen/matplotlib.animation.FuncAnimation.html#matplotlib.animation.FuncAnimation
"""
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.ticker import MaxNLocator

import time


class _AnimateBinarySearch():
    """Class to animate binary search

    Arrtibutes:
        color_list (list): list that detemines color of each bar

    Methods:
        animate_algorithm: Initialises the figure and calls the animation
        update_fig: Function called by animation.FuncAnimation to update the
                    figure

    Generators:
        color_maker: It's called by animation.FuncAnimation that yields
                     the first paramenter being used in update_fig. Yields
                     the index of where the search is and an integer respresenting
                     if we have found the number

        binary_search: Generator called by color_maker to perform binary search on
                       passed list. Yields a tuple of tuple of (int, int, int) which
                       are indices of (low, mid, high)
    """

    def __init__(self, passed_list, number, interval, title):
        """Instantiating the class.

        Args:
            passed_list (list): The list provided by the user
            number (int): The number we have to search
            interval (int): delay between frames in milliseconds
        """
        self.color_list = ['w'] * len(passed_list)

        self.animate_algorithm(passed_list, number, interval, title)

    def update_fig(self, tup, ax, passed_list, fig, low_text, mid_text, high_text):
        """Method called by animation.FuncAnimation that updates the figure for
        each frame.

        Args:
            tup (int, int, int): Tuple of color_list and tuple returned by color_maker
            ax : Matplotlib axes object
            passed_list (list): The list provided by the user
            low_text (str): setting the low text in every frame
            mid_text (str): setting the mid text in every frame
            high_text (str): setting the high text in every frame
        """

        fig.set_tight_layout(True)

        if tup[0] == -1:
            ax.bar(range(len(passed_list)), passed_list, align="edge", color=['w'] * len(passed_list))
            low_text.set_text(f"low: Not Found")
            mid_text.set_text(f"mid: Not Found")
            high_text.set_text(f"high: Not Found")
        else:
            ax.bar(range(len(passed_list)), passed_list, align="edge", color=self.color_list)
            low_text.set_text(f"low: {passed_list[tup[0]]}")
            mid_text.set_text(f"mid: {passed_list[tup[1]]}")
            high_text.set_text(f"high: {passed_list[tup[2]]}")

    def binary_search(self, arr, number):
        """Method called by color_maker to search list

        Args:
            arr (list): The list provided by the user
            number (int): The number we have to search

        Yields:
            Tuple of (int, int, int) which are indices of low, mid and high
        """
        arr = sorted(arr)
        low = 0
        high = len(arr) - 1
        mid = 0

        # the first value is never shown for some reason so
        # yielding a dummy value
        yield low, mid, high

        while low <= high:
            mid = (high + low) // 2
            yield low, mid, high

            if mid == high or mid == low:
                return

            if arr[mid] < number:
                low = mid + 1
            elif arr[mid] > number:
                high = mid - 1
            else:
                yield low, mid, high
                return
        yield -1, -1, -1

    def color_maker(self, arr, number):
        """Generator called by animation.FuncAnimation that yields the first
        paramenter being used in update_fig.

        Args:
            arr (list): The list provided by the user
            number (int): The number we have to search

        Yields:
            tup (tuple): Tuple returned by binary_search
        """
        for tup in self.binary_search(arr, number):
            low = tup[0]
            mid = tup[1]
            high = tup[2]
            self.color_list[low] = 'b'
            self.color_list[high] = 'b'
            self.color_list[mid] = 'r'
            yield tup
            self.color_list[low] = 'w'
            self.color_list[mid] = 'w'
            self.color_list[high] = 'w'

    def animate_algorithm(self, passed_list, number, interval, title):
        """Method that initializes the animation

        Args:
            passed_list (list): The list provided by the user
            number (int): The number we have to search
            interval (int): delay between frames in milliseconds
            title(str): Heading of the visualization
        """
        plt.style.use('dark_background')
        fig, ax = plt.subplots(figsize=(13, 6.5))

        fig.set_tight_layout(True)
        ax.set_xlim(0, len(passed_list))
        ax.set_ylim(0, int(1.15 * max(passed_list)))
        ax.set_title(title)

        ax.bar(range(len(passed_list)), passed_list, align="edge", color=self.color_list)

        text1 = ax.text(0.02, 0.92, "blue: max and min index \nred: mid index", transform=ax.transAxes)
        low_text = ax.text(0.02, 0.88, "low: ", transform=ax.transAxes)
        mid_text = ax.text(0.02, 0.84, "mid: ", transform=ax.transAxes)
        high_text = ax.text(0.02, 0.80, "mid: ", transform=ax.transAxes)

        ax.set_xlabel('index')
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        anim = animation.FuncAnimation(fig, func=self.update_fig,
                                       fargs=(ax, passed_list, fig, low_text, mid_text, high_text), frames=self.color_maker(passed_list, number), interval=interval,
                                       repeat=False)

        plt.show()


class _AnimateLinearSearch():
    """Class to animate Linear Search

    Attributes:
        passed_list (list): The list passed by the user

    Methods:
        animate_algorithm: Initialises the figure and calls the animation
        update_fig: Function called by animation.FuncAnimation to update the
                    figure

    Generators:
        color_maker: Generator called by animation.FuncAnimation that yields
                     the first paramenter being used in update_fig.
    """

    def __init__(self, passed_list, number, interval, title):
        """Instantiating the class.

        Args:
            passed_list (list): The list provided by the user
            number (int): The number we have to search
            interval (int): delay between frames in milliseconds
        """

        self.passed_list = passed_list
        # calling the function and starting the animation
        self.animate_algorithm(number, interval, title)

    def color_maker(self, arr, number, rect_obj_list):
        """Generator called by animation.FuncAnimation that yields the first
        paramenter being used in update_fig.

        Args:
            arr (list): The list provided by the user
            number (int): The number we have to search
            rect_obj_list (list): list of matplotlib bar containers
        Yields:
            index (int): index of where the search is
            integer : based on if we have found the element or not
        """

        # necessary for some reason, the first couple of animations
        # happen together if this statement isn't present
        yield 0, 0

        for index, num in enumerate(arr):
            if num != number:
                if index == len(arr) - 1:
                    rect_obj_list[index].set_color('b')

                    yield index, 0

                    for bar in rect_obj_list:
                        bar.set_color('w')

                    yield index, -1
                    return
                else:
                    rect_obj_list[index].set_color('b')
                    yield index, 0
            else:
                rect_obj_list[index].set_color('r')
                yield index, 1
                return

    def update_fig(self, index_, at_index, value):
        """Method called by animation.FuncAnimation that updates the figure for
        each frame.

        Args:
            index_ (int, int): The tuple returned by color_maker
            at_index (str): setting the text in every frame
            value (str): setting the text in every frame
        """

        at_index.set_text(f"at index: {index_[0]}")
        if index_[1] == -1:
            value.set_text(f"value: NOT FOUND")
        else:
            value.set_text(f"value: {self.passed_list[index_[0]]}")

    def animate_algorithm(self, number, interval, title):
        """Method that initializes the animation

        Args:
            number (int): The number we have to search
            interval (int): delay between frames in milliseconds
            title (str): the title of the animation
        """
        plt.style.use('dark_background')
        fig, ax = plt.subplots(figsize=(10, 5))
        fig.set_tight_layout(True)

        ax.set_xlim(0, len(self.passed_list))
        ax.set_ylim(0, int(1.15 * max(self.passed_list)))

        # setting every bar to white
        color_list = ['w'] * len(self.passed_list)
        rects = ax.bar(range(len(self.passed_list)), self.passed_list, align="edge", color=color_list)

        # a list of all the bar containers
        rect_obj_list = rects.get_children()

        # rect_obj_list[5].set_color
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))

        value = ax.text(0.02, 0.92, "value: ", transform=ax.transAxes)
        at_index = ax.text(0.02, 0.96, "at index: ", transform=ax.transAxes)

        anim = animation.FuncAnimation(fig, func=self.update_fig,
                                       fargs=(at_index, value), frames=self.color_maker(self.passed_list, number, rect_obj_list), interval=interval,
                                       repeat=False)

        # # code for saving the animation Future feature?
        # Writer = animation.writers['ffmpeg']
        # writer = Writer(fps=5, metadata=dict(artist='Me'), bitrate=50000)
        # anim.save(f"{title.split()[0]}.mp4", writer=writer)

        plt.show()



"""Selection sort module in sorting package.

The module is used to demonstrate the working of selection sort algorithm

Exported methods from SelectionSort class:
    sort
    evaluate
    visualize
    info
    code

Helper methods:
    __sort_it
    __time_eval_asc
    __time_eval_desc

Helper generators:
    __ascending_sort_algo
    __descending_sort_algo

Example usage:
    bs_object = sorting.SelectionSort(datalist)
    bs_object.sort(reverse = True, steps = True)
"""

import copy
import os
import sys

from ._base_sorting import BaseClass
from ._timer import Timer
from ._animation import animate_algorithm


class SelectionSort(BaseClass):
    """SelectionSort class which contains methods for analyzing selection sort algorithm.

    Attributes:
        _datalist (list): List of ints provided by the user
    """

    def __init__(self, datalist):
        """Initializes SelectionSort class with datalist.

        Args:
            datalist (list): The list provided by the user
        """
        super().__init__(datalist)
        self._datalist = datalist

    def __repr__(self):
        """__repr__ for SelectionSort class"""
        return f'algovis.sorting.selectionsort.SelectionSort({self._datalist})'

    def __ascending_sort_algo(self):
        """Helper generator for the ascending selection sort algorithm.

        It yields list after every iteration of selection sort till the
        list is sorted

        Yields:
              asc_list (list): yields complete list after each iteration of
                               the algorithm

        """
        asc_list = copy.deepcopy(self._datalist)
        length_of_list = len(asc_list)

        for i in range(length_of_list):
            # minimum unsorted value
            min_val = asc_list[i]
            min_idx = i
            for j in range(i, length_of_list):
                if asc_list[j] < min_val:
                    min_val = asc_list[j]
                    min_idx = j

            asc_list[i], asc_list[min_idx] = asc_list[min_idx], asc_list[i]
            yield asc_list

    def __descending_sort_algo(self):
        """Helper generator for the descending selection sort algorithm.

        It yields list after every iteration of selection sort till the
        list is sorted

        Yields:
              asc_list (list): yields complete list after each iteration of
                               the algorithm

        """
        asc_list = copy.deepcopy(self._datalist)
        length_of_list = len(asc_list)

        for i in range(length_of_list):
            # minimum unsorted value
            max_val = asc_list[i]
            max_idx = i
            for j in range(i, length_of_list):
                if asc_list[j] > max_val:
                    max_val = asc_list[j]
                    max_idx = j

            asc_list[i], asc_list[max_idx] = asc_list[max_idx], asc_list[i]
            yield asc_list

    def __sort_it(self, reverse, steps):
        """Helper method for 'sort' method

        It checks which generator to call based on reverse. Then it stores every
        iteration of yielded list in a dictionary. If steps is true, then _print_steps
        is called from BaseClass and the dictionary iteration_dict is passed to it which
        prints the steps

        Args:
            reverse (bool): based on reverse, we call the generator to either sort the list
                            in ascending or descending order
            steps (bool): If true, it calls function to print steps

        Returns:
            iteration_dict (dict):
        """
        iteration_dict = {}
        iterations = 0

        # the 0th iteration is basically the passed list
        iteration_dict[iterations] = self._datalist

        if not reverse:
            for yielded_list in self.__ascending_sort_algo():
                iterations += 1
                iteration_dict[iterations] = copy.deepcopy(yielded_list)
        else:
            for yielded_list in self.__descending_sort_algo():
                iterations += 1
                iteration_dict[iterations] = copy.deepcopy(yielded_list)

        if steps:
            super()._print_steps(iteration_dict, "Selection Sort")

        return iteration_dict

    def __time_eval_asc(self, iterations):
        """Helper method for 'evaluate' method

        Evaluating time of ascending selection sort algorithm.Takes in the
        'iterations' provided by the user and performs sorting that many times.

        Args:
            iterations (int): Number of times to perform sorting on this algo

        Returns:
            timing_list (list): A list of time taken for fully sorting the list
        """
        time_list = copy.deepcopy(self._datalist)
        length_of_list = len(time_list)
        timing_list = []

        while iterations:

            timer = Timer()
            timer.start()

            for i in range(length_of_list):
                # minimum unsorted value
                min_val = time_list[i]
                min_idx = i
                for j in range(i, length_of_list):
                    if time_list[j] < min_val:
                        min_val = time_list[j]
                        min_idx = j

                time_list[i], time_list[min_idx] = time_list[min_idx], time_list[i]

            stop = timer.stop()
            timing_list.append(stop)
            iterations -= 1
            time_list = copy.deepcopy(self._datalist)

        return timing_list

    def __time_eval_desc(self, iterations):
        """Helper method for 'evaluate' method

        Evaluating time of descending selection sort algorithm. Works in similar way
        to __time_eval_asc

        Args:
            iterations (int): Number of times to perform sorting on this algo

        Returns:
            timing_list (list): A list of time taken for fully sorting the list
        """
        time_list = copy.deepcopy(self._datalist)
        length_of_list = len(time_list)
        timing_list = []

        while iterations:

            timer = Timer()
            timer.start()

            for i in range(length_of_list):
                # minimum unsorted value
                max_val = time_list[i]
                max_idx = i
                for j in range(i, length_of_list):
                    if time_list[j] > max_val:
                        max_val = time_list[j]
                        max_idx = j

                time_list[i], time_list[max_idx] = time_list[max_idx], time_list[i]

            stop = timer.stop()
            timing_list.append(stop)
            iterations -= 1
            time_list = copy.deepcopy(self._datalist)

        return timing_list

    def sort(self, reverse=False, steps=False):
        """Performs selection sort on the list provided by the user.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            steps (bool): Optional; (default: False)
                            If set to True, shows iteration of each pass of
                            selection sort on the list

        Returns:
            sorted_list (list): sorted list
        """
        _sorted_object = self.__sort_it(reverse, steps)

        sorted_list = list(_sorted_object.values())[-1]

        return sorted_list

    def evaluate(self, reverse=False, iterations=1):
        """Prints the time taken to perform selection sort in nanoseconds and
        seconds to the console.

        Set optional parameter 'iterations' to the number of times you want to
        perform selection sort on the list. After every iteration, the list is reset to
        it's original unsorted state.

        Set optional parameter 'reverse' to True to sort the list in descending order.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            iterations (int): Optional; (default: 1)
                              Number of times to perform selection sort on list

        Raises:
            TypeError: When user inputs anything other than an int for iteration
        """
        if not isinstance(iterations, int):
            raise TypeError('Iterations can only be int datatype')

        if reverse:
            _timing_list = self.__time_eval_desc(iterations)
        else:
            _timing_list = self.__time_eval_asc(iterations)

        _minimum_time = min(_timing_list)
        _maximum_time = max(_timing_list)
        _average_time = int(sum(_timing_list) / iterations)

        _eval_dict = {
            "Minimum Time": _minimum_time,
            "Maximum Time": _maximum_time,
            "Average Time": _average_time
        }

        return super()._print_evaluate(_eval_dict, "Selection Sort")

    def visualize(self, reverse=False, interval=250):
        """Shows a visualization using matplotlib of selection sort performed on
        the list user passed.

        Set optional parameter 'interval' to change the delay between frames
        in milliseconds.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            interval (int): Optional; (default: 250)
                            Delay between frames in milliseconds

        Raises:
            TypeError: An error when user inputs anything other than int for
                       interval
        """
        if not isinstance(interval, int):
            raise TypeError('Interval can only be int datatype')

        _vis_list = copy.deepcopy(self._datalist)

        if not reverse:
            animate_algorithm("Selection Sort", _vis_list, self.__ascending_sort_algo(), interval)
        else:
            animate_algorithm("Selection Sort", _vis_list, self.__descending_sort_algo(), interval)

    @classmethod
    def info(cls):
        """Class method that provides information on selection sort."""
        dir = os.path.dirname(__file__)
        path_to_information = os.path.join(dir, '_markdown_files', 'selectionsort.md')
        return super()._print_info(path_to_information)

    @classmethod
    def code(cls):
        """Class method that prints code to the console."""
        my_code = """
        def selection_sort(array):
            for index, _ in enumerate(array):
                # Find the minimum element in remaining
                # unsorted array
                min_idx = index
                for j in range(index+1, len(array)):
                    if A[min_idx] > A[j]:
                        min_idx = j

                # Swap the found minimum element with
                # the first element
                A[index], A[min_idx] = A[min_idx], A[index]
        """
        return super()._print_code(my_code)



"""Quick sort module in sorting package.

The module is used to demonstrate the working of quick sort algorithm. This module
is significantly differently designed than other sorting modules currently in package,
trading readibility for complexity, although the API is same.

Exported methods from QuickSort class:
    sort
    evaluate
    visualize
    info
    code

Helper methods:
    __choose_pivot
    __sort_it
    __print_steps_quick
    __eval_helper


Helper generators:
    __quicksort

Example usage:
    qs_object = sorting.QuickSort(datalist)
    qs_object.sort(pivot = "random", reverse = True, steps = True)
"""

import copy
import random
import os
import sys

from rich.console import Console
from rich.table import Table

from ._base_sorting import BaseClass
from ._timer import Timer
from ._animation import animate_algorithm


class QuickSort(BaseClass):
    """Quick Sort class which contains methods for analyzing quick sort algorithm.

    Attributes:
        _datalist (list): List of ints provided by the user
    """

    def __init__(self, datalist):
        """Initializes Quick Sort class with datalist.

        Args:
            datalist (list): The list provided by the user
        """
        super().__init__(datalist)
        self._datalist = datalist

    def __repr__(self):
        """__repr__ for QuickSort class"""
        return f'algovis.sorting.quicksort.QuickSort({self._datalist})'

    # The next two functions are based on Lomuto partition sceheme quicksort
    # inspired by
    # https://www.geeksforgeeks.org/quicksort-using-random-pivoting/

    def __choose_pivot(self, arr, start, end, choice):
        """Helper method called by __quicksort to choose a pivot

        It chooses a pivot based on user input and what is basically does
        is that it swaps the 'pivot' with the first element and returns the
        modded array

        Args:
            arr (list): The list passed by __quicksort
            start (int): Index of first element of the array
            end (int): Index of last element of the array
            choice (str): User input
        """
        if choice == 'first':
            return arr
        elif choice == 'last':
            arr[start], arr[end] = arr[end], arr[start]
            return arr
        elif choice == 'random':
            randpivot = random.randrange(start, end)
            arr[start], arr[randpivot] = arr[randpivot], arr[start]
            return arr
        elif choice == 'middle':
            middle_pivot = start + (end - start) // 2
            arr[start], arr[middle_pivot] = arr[middle_pivot], arr[start]
            return arr
        else:
            return arr

    def __quicksort(self, arr, start, stop, choice, reverse, vis=False):
        """The helper generator called by __sort_it

        Args:
            arr (list): The list passed by __quicksort
            start (int): Index of first element of the array
            stop (int): Index of last element of the array
            choice (str): User input
            reverse (bool): If true, sorts the list in descending order
            vis (bool): Set to true for animate_algorithm, yields array at
                        different places throughout the generator

        Yields:
            arr (list): If vis == True, it yields strategically placed arrays
                        throughout the generator

            elem_pivot, arr[start:stop + 1], arr (int, list, list):
                        Yields a tuple of pivot element, list modified and total list
                        which help make the print steps

        """
        if start < stop:
            arr = self.__choose_pivot(arr, start, stop, choice)
            pivot = start
            elem_pivot = arr[pivot]  # pivot
            i = start + 1

            if vis:
                yield arr
            # a variable to memorize where the
            # partition in the array starts from.
            for j in range(start + 1, stop + 1):

                # if not reverse and the current element is smaller or equal to pivot,
                # shift it to the left side of the partition else vice versa

                if reverse:
                    if arr[j] >= arr[pivot]:
                        arr[i], arr[j] = arr[j], arr[i]
                        if vis:
                            yield arr
                        i = i + 1
                else:
                    if arr[j] <= arr[pivot]:
                        arr[i], arr[j] = arr[j], arr[i]
                        if vis:
                            yield arr
                        i = i + 1

            arr[pivot], arr[i - 1] = arr[i - 1], arr[pivot]
            pivot = i - 1

            if vis:
                yield arr
            else:
                yield elem_pivot, arr[start:stop + 1], arr
            yield from self.__quicksort(arr, start, pivot - 1, choice, reverse, vis)
            yield from self.__quicksort(arr, pivot + 1, stop, choice, reverse, vis)

    # OPTIMIZE IT by using a dictionary in next patch
    def __sort_it(self, reverse, steps, pivot, eval_=False):
        """ Function called by both, __eval_helper and __sort_it

        If the function is used for evaluation, then it just calls __quicksort
        and does nothing with returned tuple, waits for calls to finish and then
        returns None.

        If the function is called by __sort_it, then it creates a list of lists of
        every iteration and stores the data in it. If the steps == True, it calls
        __print_steps to display the evaluation on the terminal. It returns the
        iteration_list.

        Args:
        reverse (bool): bool passed to __quicksort
        steps (bool): print steps if it's true
        pivot (str): pivot chosen by the user
        eval (bool): (default: False)
        """
        if not eval_:
            passed_list = copy.deepcopy(self._datalist)

            iteration_list = []
            iterations = 0
            zeroth_iter = " ".join(str(x) for x in passed_list)
            iteration_list.append([str(0), str(" "), zeroth_iter, zeroth_iter])

            for elem_pivot, array_in_cons, array in self.__quicksort(passed_list, 0, len(passed_list) - 1, pivot, reverse):
                temp_list = []
                iterations += 1
                temp_list.append(str(iterations))
                temp_list.append(str(elem_pivot))
                temp_list.append(" ".join(str(x) for x in array_in_cons))
                temp_list.append(" ".join(str(x) for x in array))
                iteration_list.append(temp_list)

            if steps:
                self.__print_steps_quick(iteration_list)

            return iteration_list
        else:
            passed_list = copy.deepcopy(self._datalist)

            for elem_pivot, array_in_cons, array in self.__quicksort(passed_list, 0, len(passed_list) - 1, pivot, reverse):
                pass

            return

    def __print_steps_quick(self, list_of_iterations):
        """Helper method to print the list containg data about all the iterations"""
        table = Table(title="Quick Sort steps")
        table.add_column("Iteration", justify="center", style="cyan")
        table.add_column("Pivot", justify="center", style="cyan")
        table.add_column("Array in consideration", justify="center", style="magenta", no_wrap=False)
        table.add_column("Array", justify="center", style="magenta", no_wrap=False)

        for iteration in list_of_iterations:
            iter_ = str(iteration[0])
            pivot = str(iteration[1])
            array_in_cons = str(iteration[2])
            array = str(iteration[3])
            table.add_row(iter_, pivot, array_in_cons, array)

        console = Console()
        console.print(table)

    def __eval_helper(self, reverse, pivot, iterations):
        """Helper method for 'evaluate' method

        Evaluating time of ascending quick sort algorithm. Takes in the
        'iterations' provided by the user and performs sorting that many times.

        Args:
            iterations (int): Number of times to perform sorting on this algo
            reverse (bool): check if user wants to reverse the list

        Returns:
            timing_list (list): A list of time taken for fully sorting the
                                list 'iterations' number of times
        """
        timing_list = []
        steps = False
        while iterations:

            timer = Timer()
            timer.start()

            self.__sort_it(reverse, steps, pivot, eval_=True)

            stop = timer.stop()
            timing_list.append(stop)

            iterations -= 1

        return timing_list

    def sort(self, reverse=False, steps=False, pivot="first"):
        """Performs quicksort (lomuto partiton scheme) on the list
        provided by the user.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            steps (bool): Optional; (default: False)
                            If set to True, shows iteration of each pass of
                            quicksort on the list

        Returns:
            sorted_list (list): sorted list
        """
        pivot = pivot.lower()

        _sorted_object = self.__sort_it(reverse, steps, pivot)

        # the sorted list is the last list of the last element of
        # sorted obect
        return _sorted_object[-1][-1]

    def evaluate(self, reverse=False, pivot="first", iterations=1):
        """Prints the time taken to perform quick sort in nanoseconds and
        seconds to the console.

        Set optional parameter 'iterations' to the number of times you want to
        perform quick sort on the list. After every iteration, the list is reset to
        it's original unsorted state.

        Set optional parameter 'reverse' to True to sort the list in descending order.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            iterations (int): Optional; (default: 1)
                              Number of times to perform quick sort on list

        Raises:
            TypeError: When user inputs anything other than an int for iteration
        """

        if not isinstance(iterations, int):
            raise TypeError('Iterations can only be int datatype')

        pivot = pivot.lower()

        _timing_list = self.__eval_helper(reverse, pivot, iterations)

        _minimum_time = min(_timing_list)
        _maximum_time = max(_timing_list)
        _average_time = int(sum(_timing_list) / iterations)

        _eval_dict = {
            "Minimum Time": _minimum_time,
            "Maximum Time": _maximum_time,
            "Average Time": _average_time
        }

        return super()._print_evaluate(_eval_dict, "Quick Sort")

    def visualize(self, reverse=False, interval=50, pivot="first"):
        """Shows a visualization using matplotlib of quick sort performed on
        the list user passed.

        Set optional parameter 'interval' to change the delay between frames
        in milliseconds.

        Args:
            reverse (bool): Optional; (default: False)
                            If True, sorts the list in descending order
            interval (int): Optional; (default: 50)
                            Delay between frames in milliseconds

        Raises:
            TypeError: An error when user inputs anything other than int for
                       interval
        """
        if not isinstance(interval, int):
            raise TypeError('Interval can only be int datatype')

        _vis_list = copy.deepcopy(self._datalist)

        animate_algorithm("Quick Sort", _vis_list, self.__quicksort(_vis_list, 0, len(_vis_list) - 1, pivot, reverse, vis=True), interval, operations=True)

    @classmethod
    def info(cls):
        """Class method that provides information on quick sort."""
        dir = os.path.dirname(__file__)
        path_to_information = os.path.join(dir, '_markdown_files', 'quicksort.md')
        return super()._print_info(path_to_information)

    @classmethod
    def code(cls):
        my_code = """
        def partition(arr, low, high):
            # pivot
            pivot = arr[high]
            # Index of smaller element
            i = (low - 1)
            for j in range(low, high):

                # If current element is smaller than or
                # equal to pivot
                if (arr[j] <= pivot):

                    # increment index of smaller element
                    i += 1
                    arr[i], arr[j] = arr[j], arr[i]
            arr[i + 1], arr[high] = arr[high], arr[i + 1]
            return (i + 1)


        def quickSort(arr, low, high):
            if (low < high):
                pi = partition(arr, low, high)

                # Separately sort elements before
                # partition and after partition
                quickSort(arr, low, pi - 1)
                quickSort(arr, pi + 1, high)

        # To give users choice of pivot, the quicksort implemented in this library
        # is slightly different but uses lomuto partiton
        """
        return super()._print_code(my_code)
